SystemOrganization addCategory: #'MethodWrappers-Examples'!
SystemOrganization addCategory: #'MethodWrappers-Tests'!
SystemOrganization addCategory: #'MethodWrappers-Base'!

Object subclass: #MwClassA
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodWrappers-Tests'!

!MwClassA methodsFor: 'as yet unclassified' stamp: 'mp 1/22/2010 16:10'!
methodOne

	^ true! !

!MwClassA methodsFor: 'as yet unclassified' stamp: 'mp 1/22/2010 17:09'!
methodTwo: aNumber

	^ aNumber! !

Object subclass: #MwMethodWrapper
	instanceVariableNames: 'methodReference nextMethod'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodWrappers-Base'!

!MwMethodWrapper commentStamp: 'mp 1/22/2010 11:45' prior: 0!
This is a port of MwMethodWrappers by rhi for Squeak 3.11.

Method wrappers can add "hidden" behavior to the ordinary method. They can change what is executed before or after the method. For example, a count method wrapper would increment a counter when the method is called, and then forward the call onto the original method.

While this class is not abstract (i.e., there are no subclassResponsibilities), these method wrappers don't do anything if you install them. Look at the subclasses for example uses.

To create a new type of method wrapper, you must create a subclass of MwMethodWrapper and redefine the #beforeMethod and/or #afterMethod as necessary. 

To use a method wrapper, send the class a #on:inClass: message. This will return a new method wrapper (or nil if this method shouldn't be wrapped). You can then install the method wrapper by sending the #install message and uninstall by sending the #uninstall message.

WARNING: Incorrectly using the wrappers can crash your image, so you should save your image before using them. Most problems occur as a result of installing them on methods that are called either in the #beforeMethod or #afterMethod. (Note: The #value method in Association (VariableBinding in VW2.5) is automatically called by the VM when you refer to a global, class, or pool variable so you should probably never wrap it. I advise against wrapping primitives also.)

Class Instance Variables:

	protoMethods		<Dictionary
								key: SmallInteger
								value: MwMethodWrapper>
							Prototype wrappers that are copied when installing new wrappers.
	wrapperInstVars	<WeakIdentityKeyDictionary
								key: MwMethodWrapper
								value: MwMethodWrapperInstVars>
							A kludge to add instance varibles to non-pointer classes.!

MwMethodWrapper subclass: #MwAroundWrapperMethodOne
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodWrappers-Tests'!

!MwAroundWrapperMethodOne methodsFor: 'evaluating' stamp: 'mp 1/22/2010 16:52'!
aroundWithReceiver: aReceiver arguments: anArrayOfObjects

	^ false! !

MwMethodWrapper subclass: #MwAroundWrapperMethodTwo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodWrappers-Tests'!

!MwAroundWrapperMethodTwo methodsFor: '-- all --' stamp: 'mp 1/22/2010 17:15'!
aroundWithReceiver: aReceiver arguments: anArrayOfObjects

	^ (self proceed: aReceiver arguments: #(2)) + 1
! !

MwMethodWrapper subclass: #MwBeforeWrapper
	instanceVariableNames: 'touched'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodWrappers-Tests'!

!MwBeforeWrapper methodsFor: 'evaluating' stamp: 'mp 1/22/2010 15:24'!
beforeMethod

	self touched: self touched + 1.
	"Transcript show: 'X'."! !

!MwBeforeWrapper methodsFor: '-- all --' stamp: 'mp 1/22/2010 14:58'!
initialize

	self touched: 0.! !

!MwBeforeWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 14:41'!
touched
	"Answer the value of touched"

	^ touched! !

!MwBeforeWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 14:41'!
touched: anObject
	"Set the value of touched"

	touched := anObject! !

!MwMethodWrapper class methodsFor: 'testing' stamp: 'mp 1/22/2010 11:41'!
canWrap: aSelector inClass: aClass
	"Test if a method can be wrapped without causing infinite recursion."

	| class method |
	(aClass includesBehavior: MwMethodWrapper) ifTrue: [^ false].
	((aClass includesBehavior: CompiledMethod)
		and: [aSelector == #valueWithReceiver:arguments:])
			ifTrue: [^ false].
	((aClass includesBehavior: BlockContext)
		and: [aSelector == #ensure:])
			ifTrue: [^ false].
	class := aClass whichClassIncludesSelector: aSelector.
	class isNil ifTrue: [^ false].
	method := class compiledMethodAt: aSelector ifAbsent: [nil].
	^ method notNil and: [(self primitives includes: method primitive) not]! !

!MwMethodWrapper class methodsFor: 'checks' stamp: 'mp 1/22/2010 11:37'!
countAllInstances
	"doIt: [self countAllInstances]"

	| totalCount |
	totalCount := 0.
	Smalltalk garbageCollect.
	Transcript cr.
	self withAllSubclassesDo: [:each |
		| count |
		count := each allInstances size.
		totalCount := totalCount + count.
		Transcript cr; show: each name, ' -> ', count printString].
	Transcript cr; show: '=== TOTAL COUNT -> ', totalCount printString.! !

!MwMethodWrapper class methodsFor: 'private' stamp: 'mp 1/22/2010 15:33'!
nuke
	"doIt: [self nuke]"

	self	uninstallAllWrappers.! !

!MwMethodWrapper class methodsFor: 'instance creation' stamp: 'mp 1/22/2010 12:46'!
on: aSymbol inClass: aClass

	(self canWrap: aSymbol inClass: aClass) ifFalse: [^ nil].
	^ self new
		methodReference: (MethodReference class: aClass selector: aSymbol)! !

!MwMethodWrapper class methodsFor: 'private' stamp: 'mp 1/22/2010 11:41'!
primitives

	^ #(
		61		"at:put:"
		71		"new:"
		81		"value"
	)! !

!MwMethodWrapper class methodsFor: 'private' stamp: 'mp 1/22/2010 11:21'!
uninstallAllWrappers
	"doIt: [self uninstallAllWrappers]"

	self allSubInstancesDo: [:inst | inst uninstall].! !

!MwMethodWrapper methodsFor: 'private' stamp: 'mp 1/22/2010 12:34'!
afterMethod! !

!MwMethodWrapper methodsFor: 'evaluating' stamp: 'mp 1/22/2010 14:33'!
aroundWithReceiver: aReceiver arguments: anArrayOfObjects
	"This is the general case where you want both a before and after method, but if you
	want just a before method, you might want to override this method for optimization."

	self beforeMethod.
	^ [
		self proceed: aReceiver arguments: anArrayOfObjects.
	] ensure: [self afterMethod]! !

!MwMethodWrapper methodsFor: 'evaluating' stamp: 'mp 1/22/2010 12:33'!
beforeMethod! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 12:55'!
clientSelector

	^ self methodReference methodSymbol! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 17:38'!
compiledMethod

	| item |
	item := self nextMethod.
	[item isMwMethodWrapper] whileTrue: [item := item nextMethod].
	^ item! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 17:42'!
definingClass

	^ self methodReference actualClass! !

!MwMethodWrapper methodsFor: '-- all --' stamp: 'mp 1/22/2010 15:31'!
flushCache

	self class flushCache.! !

!MwMethodWrapper methodsFor: 'installation' stamp: 'mp 1/22/2010 17:46'!
install

	| methodItem |
	methodItem := self methodReference compiledMethod.
	self nextMethod: methodItem.
	self methodReference actualClass methodDictionary
		at: self methodReference methodSymbol
		put: self.! !

!MwMethodWrapper methodsFor: 'testing' stamp: 'mp 1/22/2010 12:51'!
isMwMethodWrapper

	^ true! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 11:17'!
methodReference
	"Answer the value of methodReference"

	^ methodReference! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 11:17'!
methodReference: anObject
	"Set the value of methodReference"

	methodReference := anObject! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 12:41'!
nextMethod
	"Answer the value of nextMethod"

	^ nextMethod! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 12:41'!
nextMethod: anObject
	"Set the value of nextMethod"

	nextMethod := anObject! !

!MwMethodWrapper methodsFor: 'private' stamp: 'mp 1/22/2010 14:34'!
proceed: aReceiver arguments: anArrayOfObjects

	^ self nextMethod valueWithReceiver: aReceiver arguments: anArrayOfObjects.! !

!MwMethodWrapper methodsFor: 'private' stamp: 'mp 1/22/2010 14:32'!
run: aSelector with: anArrayOfObjects in: aReceiver

	^ self aroundWithReceiver: aReceiver arguments: anArrayOfObjects.! !

!MwMethodWrapper methodsFor: 'installation' stamp: 'mp 1/22/2010 17:46'!
uninstall

	| methodItem nextItem previousItem |
	methodItem := self methodReference compiledMethod.
	
	methodItem isMwMethodWrapper ifFalse: [^ self].
	
	methodItem == self
		ifTrue: [self methodReference actualClass methodDictionary
					at: self methodReference methodSymbol	
					put: self nextMethod.
				^ self].
	
	previousItem := methodItem.
	nextItem := methodItem nextMethod.
	[nextItem isMwMethodWrapper] whileTrue: [
		nextItem == self 
			ifTrue: [	previousItem nextMethod: nextItem nextMethod.
					^ self]
			ifFalse: [	previousItem := nextItem.
					nextItem := nextItem nextMethod]].! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 17:43'!
unwrappedMethod

	^ self compiledMethod! !

!MwMethodWrapper methodsFor: 'private' stamp: 'mp 1/22/2010 14:35'!
valueWithReceiver: aReceiver arguments: anArrayOfObjects

	^ self aroundWithReceiver: aReceiver arguments: anArrayOfObjects.! !

!MwMethodWrapper methodsFor: 'accessing' stamp: 'mp 1/22/2010 17:58'!
wrappingWrapper

	| item previousItem |
	item := self methodReference compiledMethod.
	(item == self or: [item isMwMethodWrapper not]) ifTrue: [^ nil].

	previousItem := item.
	item := item nextMethod.
	[item isMwMethodWrapper] whileTrue: [
		item == self 
			ifTrue: [^ previousItem].
		previousItem := item.
		item := item nextMethod.].	
	^ nil
! !

!Object methodsFor: '*MethodWrappers' stamp: 'mp 1/22/2010 13:24'!
isMwMethodWrapper

	^ false! !

TestCase subclass: #MwWrapperTest
	instanceVariableNames: 'myObject originalMethod methodRef'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodWrappers-Tests'!

MwWrapperTest subclass: #MwAroundWrapperTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodWrappers-Tests'!

!MwAroundWrapperTest methodsFor: '-- all --' stamp: 'mp 1/22/2010 17:19'!
tearDown

	MwAroundWrapperMethodOne nuke.
	MwAroundWrapperMethodTwo nuke.
	super tearDown.! !

!MwAroundWrapperTest methodsFor: '-- all --' stamp: 'mp 1/22/2010 17:24'!
testAround

	| wrapper |
	wrapper := MwAroundWrapperMethodOne on: #methodOne inClass: MwClassA.
	self assert: self myObject methodOne.
	wrapper install.
	self deny: self myObject methodOne.
	wrapper uninstall.
	self assert: self myObject methodOne.
	
	! !

!MwAroundWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:24'!
testProceed

	| wrapper1 wrapper2 |
	wrapper1 := MwAroundWrapperMethodTwo on: #methodTwo: inClass: MwClassA.
	wrapper2 := MwAroundWrapperMethodTwo on: #methodTwo: inClass: MwClassA.	
	self assert: (self myObject methodTwo: 1) = 1.
	wrapper1 install.
	self assert: (self myObject methodTwo: 1) = 3.
	wrapper1 uninstall.
	self assert: (self myObject methodTwo: 1) = 1.
	wrapper1 install.
	wrapper2 install.
	self assert: (self myObject methodTwo: 1) = 4.
	wrapper1 uninstall.
	wrapper2 uninstall.! !

MwWrapperTest subclass: #MwBeforeWrapperTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodWrappers-Tests'!

!MwBeforeWrapperTest methodsFor: 'private' stamp: 'mp 1/22/2010 15:49'!
checkTouchs:  aWrapperAssocCollection

	aWrapperAssocCollection do: [:each | self assert: each key touched = each value].
	! !

!MwBeforeWrapperTest methodsFor: '-- all --' stamp: 'mp 1/22/2010 17:04'!
tearDown

	MwBeforeWrapper nuke.
	super tearDown.! !

!MwBeforeWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:40'!
testAccessCompiledMethod

	| wrappers |
	wrappers := OrderedCollection new.
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	wrappers first install.
	wrappers  second install.
	wrappers third install.
	self checkWrapperChain: #(Wrapper Wrapper Wrapper Original) of: self methodRef.
	self assert: wrappers third compiledMethod = originalMethod.! !

!MwBeforeWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:00'!
testInstallationMultipleWrappers

	| wrappers |
	wrappers := OrderedCollection new.
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	self checkWrapperChain: #(Original) of: self methodRef.
	wrappers first install.
	self checkWrapperChain: #(Wrapper Original) of: self methodRef.
	wrappers  second install.
	self checkWrapperChain: #(Wrapper Wrapper Original) of: self methodRef.
	self assert: self methodRef compiledMethod == wrappers second.
	wrappers  first uninstall.
	self checkWrapperChain: #(Wrapper Original) of: self methodRef.
	wrappers first install.
	self checkWrapperChain: #(Wrapper Wrapper Original) of: self methodRef.
	self assert: self methodRef compiledMethod == wrappers first.
	wrappers third install.
	self checkWrapperChain: #(Wrapper Wrapper Wrapper Original) of: self methodRef.
	self assert: self methodRef compiledMethod == wrappers third.
	self assert: self methodRef compiledMethod nextMethod == wrappers first.
	self assert: self methodRef compiledMethod nextMethod nextMethod  == wrappers second.
	wrappers first uninstall.
	self checkWrapperChain: #(Wrapper Wrapper Original) of: self methodRef.
	self assert: self methodRef compiledMethod == wrappers third.
	self assert: self methodRef compiledMethod nextMethod  == wrappers second.
	wrappers third uninstall.
	self checkWrapperChain: #( Wrapper Original) of: self methodRef.
	self assert: self methodRef compiledMethod == wrappers second.
	wrappers second uninstall.
	self checkWrapperChain: #( Original) of: self methodRef.! !

!MwBeforeWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:00'!
testInstallationOneWrapper

	| wrapper |
	wrapper := MwBeforeWrapper on: #methodOne inClass: MwClassA.
	self deny: self methodRef compiledMethod isMwMethodWrapper.
	wrapper install.
	self assert: self methodRef compiledMethod isMwMethodWrapper.
	wrapper uninstall.
	self deny: self methodRef compiledMethod isMwMethodWrapper.! !

!MwBeforeWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:00'!
testInstallationWrapperMultipleTimes

	| wrapper |
	wrapper := (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	self deny: self methodRef compiledMethod isMwMethodWrapper.
	wrapper install.
	self assert: self methodRef compiledMethod isMwMethodWrapper.
	wrapper uninstall.
	self deny: self methodRef compiledMethod isMwMethodWrapper.
	wrapper install.
	self assert: self methodRef compiledMethod isMwMethodWrapper.	
	wrapper uninstall.! !

!MwBeforeWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:02'!
testOneWrapper

	| wrapper |
	wrapper := MwBeforeWrapper on: #methodOne inClass: MwClassA.
	self assert: self myObject methodOne.
	self checkTouchs: {wrapper -> 0}. 
	wrapper install.
	self checkTouchs: {wrapper -> 0}. 
	self assert: self myObject methodOne.
	self checkTouchs: {wrapper -> 1}. 
	wrapper uninstall.
	self assert: ((MethodReference class: MwClassA selector: #methodOne) compiledMethod isKindOf: CompiledMethod).
	self checkTouchs: {wrapper -> 1}. 
	self assert: self myObject methodOne.
	self checkTouchs: {wrapper -> 1}.! !

!MwBeforeWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:02'!
testTwoWrappers

	| wrapper1 wrapper2 |
	wrapper1 := MwBeforeWrapper on: #methodOne inClass: MwClassA.
	wrapper2 := MwBeforeWrapper on: #methodOne inClass: MwClassA.
	self checkTouchs: {wrapper1 -> 0. wrapper2 -> 0}. 
	self assert: self myObject methodOne.
	self checkTouchs: {wrapper1 -> 0. wrapper2 -> 0}. 	
	wrapper1 install.
	self assert: self myObject methodOne.
	self checkTouchs: {wrapper1 -> 1. wrapper2 -> 0}. 
	wrapper2 install.
	self assert: self myObject methodOne.
	self checkTouchs: {wrapper1 -> 2. wrapper2 -> 1}. 	
	wrapper2 uninstall.
	self assert: self methodRef compiledMethod isMwMethodWrapper.
	self assert: self myObject methodOne.
	self checkTouchs: {wrapper1 -> 3. wrapper2 -> 1}. 
	wrapper1 uninstall.
	self deny:  self methodRef compiledMethod isMwMethodWrapper.
	self assert: self myObject methodOne.
	self checkTouchs: {wrapper1 -> 3. wrapper2 -> 1}. 	! !

!MwBeforeWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:02'!
testWithoutWrapper

	self assert: self myObject methodOne.! !

!MwBeforeWrapperTest methodsFor: 'tests' stamp: 'mp 1/22/2010 17:57'!
testWrappingWrapper

	| wrappers |
	wrappers := OrderedCollection new.
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	wrappers add: (MwBeforeWrapper on: #methodOne inClass: MwClassA).
	wrappers first install.
	wrappers  second install.
	wrappers third install.
	self checkWrapperChain: #(Wrapper Wrapper Wrapper Original) of: self methodRef.
	self assert: wrappers third wrappingWrapper = nil.
	self assert: wrappers second wrappingWrapper = wrappers third.
	self assert: wrappers first wrappingWrapper = wrappers second.
	wrappers second uninstall.
	self assert: wrappers second wrappingWrapper = nil.
	self assert: wrappers first wrappingWrapper = wrappers third.! !

!MwWrapperTest methodsFor: 'private' stamp: 'mp 1/22/2010 16:53'!
checkWrapperChain: anArray of: aMethodRef

	| item |
	item := aMethodRef compiledMethod.
	anArray do: [:each |
		each == #Wrapper
			ifTrue: [	self assert: item isMwMethodWrapper.
					item := item nextMethod].
		each == #Original
			ifTrue: [self deny: item isMwMethodWrapper]].! !

!MwWrapperTest methodsFor: 'accessing' stamp: 'mp 1/22/2010 16:54'!
methodRef
	"Answer the value of methodRef"

	^ methodRef! !

!MwWrapperTest methodsFor: 'accessing' stamp: 'mp 1/22/2010 16:54'!
methodRef: anObject
	"Set the value of methodRef"

	methodRef := anObject! !

!MwWrapperTest methodsFor: 'accessing' stamp: 'mp 1/22/2010 16:55'!
myObject
	"Answer the value of myObject"

	^ myObject! !

!MwWrapperTest methodsFor: 'accessing' stamp: 'mp 1/22/2010 16:55'!
myObject: anObject
	"Set the value of myObject"

	myObject := anObject! !

!MwWrapperTest methodsFor: '-- all --' stamp: 'mp 1/22/2010 16:56'!
setUp

	self myObject: MwClassA new.
	originalMethod := (MethodReference class: MwClassA selector: #methodOne) compiledMethod.
	self methodRef: (MethodReference class: MwClassA selector: #methodOne).! !

!MwWrapperTest methodsFor: '-- all --' stamp: 'mp 1/22/2010 16:56'!
tearDown

	self methodRef actualClass methodDictionary at: self methodRef methodSymbol put: originalMethod.! !
